import { singleton } from 'decorators/singleton'
import { TaskType } from 'models/RequestTask'
import { ConfirmQueue } from 'queue/confirm.queue'
import Web3 from 'web3'
import { TransactionReceipt, AddedAccount } from 'web3-core'
import { ERC20Reactor } from './ERC20Reactor'
import { ERC721Reactor } from './ERC721Reactor'
import { HttpRetryProvider } from './HttpRetryProvider'
import { WalletReactor } from './WalletReactor'
import { DistributorReactor } from './DistributorReactor'
import { fromTokenMinimalUnit, safeNumberToBN, toBN } from 'utils/number.util'
import { AllChains } from './allchain'
import assert from 'assert'
import { IPriceData } from 'structs/PriceData'
import { IChainData } from 'structs/ChainData'
import { PriceSvr } from 'service/price.service'
import { ActivityReactor } from './ActivityReactor'

@singleton
export class BlockChain {
  public web3: Web3
  instanceCacheMap: Map<string, any>
  private accountMaster: AddedAccount
  private currentChain: IChainData
  public erc20Reactor: ERC20Reactor
  public erc721Reactor: ERC721Reactor
  public walletReactor: WalletReactor
  public activityReactor: ActivityReactor
  public distributorReactor: DistributorReactor
  public confirmQueue: ConfirmQueue
  public currentBlockNum: number = 0

  constructor() {
    const defaultChain = parseInt(process.env.CHAIN_DEFAULT)
    this.currentChain = AllChains.find(o => o.id === defaultChain)
    assert(this.currentChain, `chain data with ${defaultChain} not found`)
    const provider = new HttpRetryProvider(this.currentChain.rpc.split('|'))
    this.web3 = new Web3(provider)
    this.web3.eth.handleRevert = true
    this.confirmQueue = new ConfirmQueue(this.web3)
    let key = process.env.CHAIN_MASTER_KEY
    this.accountMaster = this.web3.eth.accounts.wallet.add(key)
    this.instanceCacheMap = new Map()
    this.erc20Reactor = new ERC20Reactor({
      web3: this.web3,
      address: process.env.CHAIN_FT_ADDRESS,
    })
    this.erc721Reactor = new ERC721Reactor({
      web3: this.web3,
      address: process.env.CHAIN_NFT_ADDRESS,
    })
    this.walletReactor = new WalletReactor({
      web3: this.web3,
      address: process.env.CHAIN_WALLET_ADDRESS,
    })
    this.distributorReactor = new DistributorReactor({
      web3: this.web3,
      address: process.env.CHAIN_DISTRIBUTOR_ADDRESS,
    })
    this.activityReactor = new ActivityReactor({
      web3: this.web3,
      address: process.env.CHAIN_CLAIM_ADDRESS,
    })
  }

  public get currentAccount() {
    return this.accountMaster.address
  }

  public async getContractInstance(address: string, abi: any) {
    if (!this.instanceCacheMap.has(address)) {
      const instance = new this.web3.eth.Contract(abi, address, { from: this.accountMaster.address })
      this.instanceCacheMap.set(address, instance)
    }
    return this.instanceCacheMap.get(address)
  }

  public async getTransactionReceipt(txHash: string) {
    return this.web3.eth.getTransactionReceipt(txHash)
  }

  public async getTxConfirms(txhash: string) {
    const receipt: TransactionReceipt = await this.getTransactionReceipt(txhash)
    const latest = await this.web3.eth.getBlockNumber()
    return latest - receipt.blockNumber + 1
  }

  public async updateCurrenBlockNum() {
    this.currentBlockNum = await this.web3.eth.getBlockNumber()
    // logger.debug(`update block num: ${this.currentBlockNum}`)
  }

  public async generateFunAbi(reqData: any) {
    let taskType = parseInt(reqData.type)
    reqData.encodeABI = true
    let abi
    switch (taskType) {
      case TaskType.MINT_FT:
        abi = await this.erc20Reactor.mint(reqData)
        break
      case TaskType.MINT_NFT:
        reqData.tokenId = reqData.tokenId || reqData.tokenid
        abi = await this.erc721Reactor.mint(reqData)
        break
      case TaskType.TRANSFER_FT:
        abi = await this.erc20Reactor.transfer(reqData)
        break
      case TaskType.TRANSFER_NFT:
        reqData.tokenId = reqData.tokenId || reqData.tokenid
        abi = await this.erc721Reactor.transfer(reqData)
        break
      case TaskType.PUBLISH_AIRDROP_LIST:
        reqData.nftList = [reqData.tokenId || reqData.tokenid]
        abi = await this.distributorReactor.publishAirdropList(reqData)
        break
    }
    return abi
  }

  public async sendEth({ from, to, amount }: { from: string; to: string; amount: number }) {
    const amountToSend = this.web3.utils.toWei(amount + '', 'ether')
    let gas = await this.web3.eth.estimateGas({
      from,
      to,
      value: '0',
    })
    return this.web3.eth.sendTransaction({
      from,
      to,
      gas: 21000,
      value: amountToSend,
    })
  }

  public async generateGasShow(gas: any): Promise<IPriceData> {
    let price = await new PriceSvr().refreshGasPrice()
    let ehtBN = safeNumberToBN(price).mul(safeNumberToBN(gas))
    let ethSymbol = this.currentChain.type !== 'Testnet' ? this.currentChain.symbol : 'ETH'
    let leagelPrice = await new PriceSvr().queryEthPrice(ethSymbol)
    let leagelPriceBN = safeNumberToBN(leagelPrice)
    let leagel = fromTokenMinimalUnit(ehtBN.mul(leagelPriceBN), 20)
    let eth = fromTokenMinimalUnit(ehtBN, 18)
    return { gas, price, eth, leagel }
  }
}
