import { BN } from 'ethereumjs-util'
import Web3 from 'web3'

import { Contract } from 'web3-eth-contract'
import { Account } from 'web3-core'
import { AllChains } from './allchain'
import { HttpRetryProvider } from './HttpRetryProvider'

const abiFt = require('abis/ERC20.json').abi
export class ERC20Reactor {
  private web3: Web3
  private contract: Contract
  private account: Account

  constructor({ web3, address }: { web3: Web3; address: string }) {
    this.web3 = web3
    this.account = this.web3.eth.accounts.wallet[0]
    this.contract = new this.web3.eth.Contract(abiFt, address, { from: this.account.address })
  }

  /**
   * Get balance or count for current account on specific asset contract.
   *
   * @param address - Asset ERC20 contract address.
   * @param selectedAddress - Current account public address.
   * @returns Promise resolving to BN object containing balance for current account on specific asset contract.
   */
  async getBalanceOf({ address, selectedAddress }: { address?: string; selectedAddress: string }): Promise<BN> {
    const contract = address
      ? new this.web3.eth.Contract(abiFt, address, { from: this.account.address })
      : this.contract
    return new Promise<BN>((resolve, reject) => {
      contract.methods.balanceOf(selectedAddress).call({ from: selectedAddress }, (error: Error, result: BN) => {
        /* istanbul ignore if */
        if (error) {
          reject(error)
          return
        }
        resolve(result)
      })
    })
  }

  /**
   * Query for the decimals for a given ERC20 asset.
   *
   * @param address - ERC20 asset contract string.
   * @returns Promise resolving to the 'decimals'.
   */
  async getTokenDecimals(address?: string): Promise<string> {
    const contract = address
      ? new this.web3.eth.Contract(abiFt, address, { from: this.account.address })
      : this.contract
    return new Promise<string>((resolve, reject) => {
      contract.methods.decimals().call((error: Error, result: BN | string) => {
        /* istanbul ignore if */
        if (error) {
          reject(error)
          return
        }
        resolve(result.toString())
      })
    })
  }

  /**
   * Query for symbol for a given ERC20 asset.
   *
   * @param address - ERC20 asset contract address.
   * @returns Promise resolving to the 'symbol'.
   */
  async getTokenSymbol(address?: string): Promise<string> {
    const contract = address
      ? new this.web3.eth.Contract(abiFt, address, { from: this.account.address })
      : this.contract
    return new Promise<string>((resolve, reject) => {
      contract.methods.symbol().call((error: Error, result: BN | string) => {
        /* istanbul ignore if */
        if (error) {
          reject(error)
          return
        }
        resolve(result.toString())
      })
    })
  }
  /**
   * Query if a contract implements an interface.
   *
   * @param address - Asset contract address.
   * @param userAddress - The public address for the currently active user's account.
   * @returns Promise resolving an object containing the standard, decimals, symbol and balance of the given contract/userAddress pair.
   */
  async getDetails({ address, userAddress }: { address?: string; userAddress: string }): Promise<{
    standard: string
    symbol: string | undefined
    decimals: string | undefined
    balance: BN | undefined
  }> {
    const [decimals, symbol] = await Promise.all([this.getTokenDecimals(address), this.getTokenSymbol(address)])
    let balance
    if (userAddress) {
      balance = await this.getBalanceOf({ address, selectedAddress: userAddress })
    }
    return {
      decimals,
      symbol,
      balance,
      standard: 'ERC20',
    }
  }

  async transfer({
    address,
    from,
    to,
    amount,
    account,
    chain,
    gas,
    encodeABI = false,
    estimate = false,
  }: {
    address: string
    from?: string
    to: string
    account?: string
    chain?: number
    amount: number | string
    gas?: number
    encodeABI?: boolean
    estimate?: boolean
  }) {
    from = from || account || this.account.address
    let w3 = this.web3
    if (chain) {
      let chainData = AllChains.find(o => o.id === chain)
      if (chainData) {
        const provider = new HttpRetryProvider(chainData.rpc.split('|'))
        w3 = new Web3(provider)
      }
    }
    const contract = new w3.eth.Contract(abiFt, address, { from: account || this.account.address })
    const amountBN = Web3.utils.toBN(Web3.utils.toWei(amount + ''))
    if (encodeABI) {
      if (from === this.account.address) return contract.methods.transfer(to, amountBN).encodeABI()
      return contract.methods.transferFrom(from, to, amountBN).encodeABI()
    }
    let gasEstimate =
      from === this.account.address
        ? await contract.methods.transfer(to, amountBN).estimateGas()
        : await contract.methods.transferFrom(from, to, amountBN).estimateGas()
    if (estimate) {
      return gasEstimate
    }
    if (from === this.account.address) return contract.methods.transfer(to, amountBN).send({ gas: gas || gasEstimate })
    return contract.methods.transferFrom(from, to, amountBN).send({
      gas: gas || gasEstimate,
    })
  }

  async mint({
    address,
    to,
    amount,
    account,
    chain,
    encodeABI = false,
  }: {
    account?: string
    address?: string
    to: string
    amount: string
    chain?: number
    encodeABI: boolean
  }) {
    let w3 = this.web3
    if (chain) {
      let chainData = AllChains.find(o => o.id === chain)
      if (chainData) {
        const provider = new HttpRetryProvider(chainData.rpc.split('|'))
        w3 = new Web3(provider)
      }
    }
    const contract = address
      ? new w3.eth.Contract(abiFt, address, { from: account || this.account.address })
      : this.contract
    let amountBN = Web3.utils.toBN(Web3.utils.toWei(amount + ''))
    if (encodeABI) {
      return contract.methods.mint(to, amountBN).encodeABI()
    }
    let gas = await contract.methods.mint(to, amountBN).estimateGas({ gas: 1000000 })
    return contract.methods.mint(to, amountBN).send({ gas: (gas * 1.1) | 0 })
  }

  async getPastEvents({ address, fromBlock }: { address?: string; fromBlock: number }) {
    const contract = address
      ? new this.web3.eth.Contract(abiFt, address, { from: this.account.address })
      : this.contract
    return contract.getPastEvents('Transfer', {
      fromBlock,
      toBlock: fromBlock + 50000,
    })
  }
}
