import axios from 'axios'
import deepmerge from 'deepmerge'
import logger from 'logger/logger'
import { generateHeader } from 'utils/net.util'
import { retry } from 'utils/promise.util'

const defaultOptions = {
  retry: {
    retries: 5,
  },
}

export class HttpRetryProvider {
  public currentIndex: number
  public options: any
  public urls: string[]

  constructor(urls: string[], options: any = {}) {
    if (!urls || !urls.length) {
      throw new TypeError(`Invalid URLs: '${urls}'`)
    }
    this.urls = urls
    this.options = deepmerge(defaultOptions, options)
    this.currentIndex = 0
  }

  public async send(payload, callback) {
    try {
      const [result, index] = await retry(() => trySend(payload, this.urls, this.currentIndex), this.options.retry)
      this.currentIndex = index
      callback(null, result)
    } catch (e) {
      callback(e)
    }
  }

  public async sendAsync(payload, callback) {
    try {
      const [result, index] = await retry(() => trySend(payload, this.urls, this.currentIndex), this.options.retry)
      this.currentIndex = index
      callback(null, result)
    } catch (e) {
      callback(e)
    }
  }
}

export class HttpRetryProviderError extends Error {
  public errors: any
  constructor(message, errors) {
    super(message)
    this.errors = errors
  }
}

async function trySend(payload, urls, initialIndex) {
  const errors = []

  let index = initialIndex
  for (let count = 0; count < urls.length; count++) {
    const url = urls[index]
    const headers = generateHeader()
    try {
      const config: any = {
        method: 'post',
        url,
        headers,
        data: JSON.stringify(payload),
        timeout: 30 * 1000,
      }
      const response = await axios(config)
      if (response.status !== 200) {
        const text = response.data
        throw new Error(`[${response.status}: ${url}]: ${text}`)
      }
      const result = response.data
      return [result, index]
    } catch (e) {
      // log error here
      logger.error({ error: e.message }, 'fetch fail')
      errors.push(e)
    }
    index = (index + 1) % urls.length
  }

  throw new HttpRetryProviderError('Request failed for all urls', errors)
}
