import axios from 'axios'
import logger from 'logger/logger'
import { RedisClient } from 'redis/RedisClient'

export function queryNftTxCount(address: string) {
  let data = {
    method: 'Rocket_queryTxCount',
    jsonrpc: '2.0',
    id: 1,
    params: ['token_transfer_tx', address],
  }
  let url = `${process.env.EXPLOR_RPC_URL}`
  let reqConfig: any = {
    method: 'post',
    url,
    headers: {
      'Content-Type': 'application/json',
    },
    data: JSON.stringify(data),
  }
  return axios(reqConfig).then(res => res.data)
}

export function queryNftTx(address: string, page: number, limit: number = 100) {
  let data = {
    method: 'Rocket_pageQueryTx',
    jsonrpc: '2.0',
    id: 1,
    params: ['token_transfer_tx', page, limit, address],
  }
  let url = `${process.env.EXPLOR_RPC_URL}`
  let reqConfig: any = {
    method: 'post',
    url,
    headers: {
      'Content-Type': 'application/json',
    },
    data: JSON.stringify(data),
  }
  return axios(reqConfig).then(res => res.data)
}

export class ExploreNftTxSvr {
  address: string
  lastBlock: number
  blockKey: string
  eventProcesser: any

  constructor({
    address,
    event,
    abi,
    fromBlock,
    eventProcesser,
  }: {
    address: string
    event: string
    abi: any
    fromBlock: number
    eventProcesser: any
  }) {
    this.address = address.toLowerCase()
    this.lastBlock = 0
    this.blockKey = `${address.toLowerCase()}_Transfer`
    this.eventProcesser = eventProcesser
  }

  async parseEvents() {
    logger.info(`query nft tx list:: address: ${this.address}`)
    let blockStr = await new RedisClient().get(this.blockKey)
    if (blockStr) {
      this.lastBlock = Math.max(parseInt(blockStr), this.lastBlock)
    }

    const pageSize = this.lastBlock > 0 ? 25 : 100
    let page = 0
    let minBlock = this.lastBlock
    let maxBlock = 0
    let next = true
    while (next) {
      logger.info(`parse tx event for : ${this.address} page: ${page}`)
      let result
      try {
        result = await queryNftTx(this.address, page, pageSize)
      } catch (err) {
        logger.info(`error parse tx event for : ${this.address} page: ${page}`)
        await new Promise(resolve => setTimeout(resolve, 1000))
        continue
      }

      const { status, data } = result.result
      if (!!status) {
        // 请求出错, 暂停一秒重试
        await new Promise(resolve => setTimeout(resolve, 1000))
        continue
      }
      if (data.length === 0) {
        // 没数据了
        next = false
        continue
      }
      for (let event of data) {
        maxBlock = Math.max(maxBlock, event.blockHeight)
        if (minBlock > 0) {
          minBlock = Math.min(event.blockHeight, minBlock)
        } else {
          minBlock = event.blockHeight
        }
        await this.eventProcesser.saveEvent(event)
      }
      // 如果当前批次最小的blockHeight 小于上次处理的最大blockHeight
      // 那么说明这批次里已经有上次已处理的事件, 那么可以跳出循环,结束当前查询了
      if (minBlock <= this.lastBlock) {
        next = false
        continue
      }
      page += 1
    }
    new RedisClient().set(this.blockKey, maxBlock + '')
  }
}
