import { BlockChain } from 'chain/BlockChain'
import { singleton } from 'decorators/singleton'
import * as schedule from 'node-schedule'
import logger from 'logger/logger'
import axios from 'axios'

export interface IPriceCache {
  price: string
  expired: number
}
@singleton
export class PriceSvr {
  private priceMap: Map<string, IPriceCache> = new Map()

  public async refreshGasPrice(refresh: boolean = false) {
    const key = `gasprice`
    let data = this.priceMap.get(key)
    if (!data || data.expired < Date.now() || refresh) {
      let price = await new BlockChain().web3.eth.getGasPrice()
      this.priceMap.set(key, { price, expired: Date.now() + 1000 * 60 })
    }
    return this.priceMap.get(key).price
  }

  public async queryEthPrice(eth: string, refresh: boolean = false) {
    const usd = 'USD'
    const key = `crypto_usd|${eth}`
    let data = this.priceMap.get(key)
    if (!data || data.expired < Date.now() || refresh) {
      const key = process.env.CRYPTOCOMPARE_API_KEY
      const url = `https://min-api.cryptocompare.com/data/price?fsym=${eth}&tsyms=${usd}&api_key=${key}`
      let priceData = await axios.get(url).then(res => res.data)
      let price = priceData[usd] * 100 + ''
      this.priceMap.set(key, { price, expired: Date.now() + 1000 * 60 })
    }
    return this.priceMap.get(key).price
  }

  private async refreshAll() {
    for (let key of this.priceMap.keys()) {
      let [type, token] = key.split('|')
      if (type == 'gasprice') {
        try {
          await this.refreshGasPrice(true)
        } catch (e) {
          logger.info(`error refresh gas price `, e.message || e)
        }
      } else if (type == 'crypto_usd') {
        try {
          await this.queryEthPrice(token, true)
        } catch (e) {
          logger.info(`error refresh crypto price: ${token}`, e.message || e)
        }
      }
    }
  }
  scheduleAll() {
    const job = schedule.scheduleJob('*/30 * * * * *', async () => {
      this.refreshAll()
    })
  }
}
