import logger from 'logger/logger'
import { toBN } from './number.util'
import { BN } from 'ethereumjs-util'
import { RedisClient } from 'redis/RedisClient'

const ONE = toBN(1)
const TWO = toBN(2)
const queryRange = toBN(1000)
// 返回数据如果达到这个数值, 需要拆分块的区间, 重新获取
const RESULT_LIMIT_COUNT = parseInt(process.env.RESULT_LIMIT_COUNT)
// 单个块event数量超过该值, 需要独立请求
const SPLIT_LIMIT_COUNT = 40

const blockTimeMap: Map<number, number> = new Map()

async function divQueryPassEvents({
  contract,
  event,
  fromBlock,
  toBlock,
  options,
}: {
  contract: any
  event: string
  fromBlock: BN
  toBlock: BN
  options?: any
}) {
  const middle = fromBlock.add(toBlock).divRound(TWO)
  const middlePlusOne = middle.add(ONE)

  const firstHalfEvents = await getPastEvents({
    contract,
    event,
    fromBlock,
    toBlock: middle,
    options,
  })
  const secondHalfEvents = await getPastEvents({
    contract,
    event,
    fromBlock: middlePlusOne,
    toBlock,
    options,
  })
  return [...firstHalfEvents, ...secondHalfEvents]
}

export async function getPastEvents({
  contract,
  event,
  fromBlock,
  toBlock,
  options,
}: {
  contract: any
  event: string
  fromBlock: BN
  toBlock: BN
  options?: any
}) {
  logger.debug(`${contract.options.address}: ${event} from: ${fromBlock} to: ${toBlock}`)
  let events
  try {
    events = await contract.getPastEvents(event, {
      ...options,
      fromBlock,
      toBlock,
    })
    logger.debug(`from: ${fromBlock} to: ${toBlock} events: ${events.length}`)
    // 有些公共节点会限制返回事件数量, 需要拆分块的区间, 重新获取
    if (RESULT_LIMIT_COUNT > 0 && events.length >= RESULT_LIMIT_COUNT) {
      events = await divQueryPassEvents({
        contract,
        event,
        fromBlock,
        toBlock,
        options,
      })
    }
  } catch (e) {
    if (
      e.message &&
      (/query returned more than \d+ results/.test(e.message) ||
        /logs matched by query exceeds limit of \d/.test(e.message))
    ) {
      events = await divQueryPassEvents({
        contract,
        event,
        fromBlock,
        toBlock,
        options,
      })
    } else {
      throw new Error(e)
    }
  }
  return events
}

export function* getPastEventsIter({
  chain,
  contract,
  event,
  fromBlock,
  toBlock,
  options,
}: {
  chain: number
  contract: any
  event: string
  fromBlock: number
  toBlock: number
  options?: any
}) {
  const address = contract.options.address
  const redisKey = `${chain}_${address.toLowerCase()}_${event}`
  logger.debug(`*getPastEventsIter: ${event} from: ${fromBlock} to: ${toBlock}`)
  let from = toBN(fromBlock)
  let to = toBN(fromBlock).add(queryRange)
  const toBlockBN = toBN(toBlock)
  while (to.lt(toBlockBN)) {
    yield getPastEvents({ contract, event, fromBlock: from, toBlock: to, options })
    from = to.add(ONE)
    to = to.add(queryRange)
    yield new RedisClient().set(redisKey, from + '')
  }
  yield getPastEvents({ contract, event, fromBlock: from, toBlock: toBlockBN, options })
  yield new RedisClient().set(redisKey, toBlockBN.add(ONE) + '')
}

export async function processEvents(web3, iterator, chain: number, processedEvent) {
  for (const getPastEventPromise of iterator) {
    const events = await getPastEventPromise
    for (const event of events) {
      if (event?.blockNumber) {
        if (blockTimeMap.has(event.blockNumber)) {
          event.timestamp = blockTimeMap.get(event.blockNumber)
        } else {
          const blockData = await web3.eth.getBlock(event.blockNumber)
          event.timestamp = blockData.timestamp
          blockTimeMap.set(event.blockNumber, blockData.timestamp)
        }
      }
      event.chain = chain
      await processedEvent(event)
    }
  }
}

export function clearTimeCache() {
  blockTimeMap.clear()
}
